import sys
import os

if __package__ and "rook" in __package__:
    from rook.lib.config import ConfigManager, ConfigurationScheme
else:
    from lib.config import ConfigManager, ConfigurationScheme

try:
    from .certificates import ROOKOUT_SERVER_CERT
except ImportError:
    # This is caused by not changing sys.path in Rook combined with the way we package the code
    ROOKOUT_SERVER_CERT = None


########################################################################################################################
# General Purpose Configurations
########################################################################################################################


class LoggingConfiguration(ConfigurationScheme):
    LOGGER_NAME = None
    FILE_NAME = None
    LOG_TO_STDERR = True
    LOG_LEVEL = "DEBUG"
    PROPAGATE_LOGS = False
    LOG_TO_REMOTE = False


class VersionConfiguration(ConfigurationScheme):
    VERSION = "No Version Info"
    COMMIT = "No Commit Info"


class CertificatesConfig(ConfigurationScheme):
    # True ( Verify with signed CA) | False (Don't verify) | "path/crt.crt" Use specific certificate
    VERIFY_SERVER_SSL = True


class DefaultConfiguration(ConfigurationScheme):
    DEFAULT_CONFIG_FILE_NAME = "rookout-config.json"
    LOAD_FROM_APP_FOLDER = True
    IGNORE_FILE = None
    USER_CONFIGURATION_FILE = None


class ErrorConfiguration(ConfigurationScheme):
    STACK_DEPTH = 10


class ProtobufConfiguration(ConfigurationScheme):
    NAMESPACE_SERIALIZER_DUMPING = r""


########################################################################################################################
# Rook Configurations
########################################################################################################################


class AgentAddress(ConfigurationScheme):
    HOST = 'localhost'
    PORT = 7486


class AgentComConfiguration(ConfigurationScheme):
    COMMAND_THREAD_NAME = "rookout_agent_com"
    MAX_MESSAGE_LENGTH = 100 * 1024 * 1024
    TOTAL = 2 ** 20
    CONNECT = TOTAL
    READ = 3
    BACK_OFF = 0.2
    MAX_SLEEP = 60
    GRPC_TIMEOUT = 5


class OutputConfiguration(ConfigurationScheme):
    FLUSH_TIME_INTERVAL = 0.25
    THREAD_NAME = "rookout_output_thread"
    MAX_ITEMS = 40
    MAX_LOG_ITEMS = 100
    MAX_STATUS_UPDATES = 100


class InstrumentationConfig(ConfigurationScheme):
    ENGINE = "auto"


class ImportServiceConfig(ConfigurationScheme):
    USE_IMPORT_HOOK = False
    SYS_MODULES_QUERY_INTERVAL = 0.25
    THREAD_NAME = "rookout_import_service_thread"


class LoggingServiceConfig(ConfigurationScheme):
    BASIC_CONFIG_ROOT = True


########################################################################################################################
# Agent Configurations
########################################################################################################################


class ConfigurationFiles(ConfigurationScheme):
    AGENT_IDENTITY = "identity.json"


class RookServerConfiguration(ConfigurationScheme):
    LISTEN_LOCAL = True
    PORT = 7486
    THREADS = 150
    MAX_MESSAGE_LENGTH = 30 * 1024 * 1024
    MAX_RPC_ERRORS = 5
    SECURE = False


class HttpRetries(ConfigurationScheme):
    TOTAL = 2 ** 20
    CONNECT = TOTAL
    READ = 0
    BACK_OFF = 0.2
    MAX_SLEEP = 120


class RooksManagerConfiguration(ConfigurationScheme):
    MAX_DELAY = 3


class LoginInformation(ConfigurationScheme):
    TOKEN = "7f8c7602f02fcb0d9feaf15bc19adeadbeefb4b32c637d8c4b037b4d558e7bdd"


class ControlServerConfiguration(ConfigurationScheme):
    HOST = "app.rookout.com"
    SECURE = True

    MAX_RETRIES = 1024

    DEFAULT_KEYWORDS = {'org_id': 'rookout'}
    USER = "service"

    THREADS = 4

    SEND_DATA = True

    class WebSocket(ConfigurationScheme):
        URL = "/apiv1/rtm/agents"

    class Rest(ConfigurationScheme):
        pass


class ProcessingConfiguration(ConfigurationScheme):
    RULE_REMOVAL_DELAY = 300
    RULE_DELETION_FREQUENCY = 10
    THREAD_POOL_SIZE = 2
    MAX_THREAD_POOL_DEPTH = 80


class AgentOutputConfiguration(ConfigurationScheme):
    THREAD_POOL_SIZE = 2
    FLUSH_TIME_INTERVAL = 1
    THREAD_NAME = "agent_output_thread"

class HttpHealthCheck(ConfigurationScheme):
    PORT = None

class AgentIdentity(ConfigurationScheme):
    ID = None


class MachineIdentity(ConfigurationScheme):
    ID = None


class AgentInfo(ConfigurationScheme):
    TAGS = []


########################################################################################################################
# Backend Configurations
########################################################################################################################


class RookoutServerConfig(ConfigurationScheme):
    # Port server will listen on
    PORT = 443
    # Key used by server to encrypt cookies
    SECRET_KEY = None
    # Admin user default password
    ADMIN_TUPLE = None
    # The name agents used for token based logins
    AGENT_USER_NAME = "service"
    # Service user default password
    SERVICES_USER_TUPLE = None
    # SSL {"certfile": "file.crt", "keyfile": "file.key"} (HTTPS) | None (HTTP instead of HTTPS)
    SSL_OPTIONS = None
    # The REDIS SERVER used for DB "104.197.26.251"  # Internal="10.128.0.3"  or None for localhost
    DB_HOST = None
    # Additional ports the server will listen on and will redirect to chosen port {port: (port, scheme)}
    REDIRECT_PORTS = {}
    # Should the server run in test mode
    TEST_MODE = False
    # Values used to create a test organization
    TEST_ORGANIZATION_INFO = {}
    TEST_USER_INFO = {}
    TEST_ADMIN_INFO = {}
    TEST_LICENSE = {}


class BackendDbConfig(ConfigurationScheme):
    DB_TYPE = None
    DB_FACTORY = None
    DB_NAME = 'mymaster'


class BackendAssetsLimits(ConfigurationScheme):
    RULE_STATUS_INDEX_LIMIT = 256
    RULE_STATUS_DELETE = 32


class ExternalServices(ConfigurationScheme):
    class Github(ConfigurationScheme):
        NAME = "github"
        CLIENT_ID = None
        CLIENT_SECRET = None

    class Loggly(ConfigurationScheme):
        NAME = "loggly"
        TOKEN = None


manager = ConfigManager(sys.modules[__name__])


def _load_config_file():
    # Calculate package path
    if 'rook' in __name__:
        # Find the rook directory
        import rook
        package_path = rook.__path__[0]
    elif sys.argv[0].endswith('/rookout-agent.py') or sys.argv[0].endswith('rookout-agent'):
        try:
            import agent
            package_path = agent.__path__[0]
        except ImportError:
            pass
    else:
        package_path = None

    # If a package configuration file exists, load it
    if package_path:
        package_config = os.path.join(package_path, DefaultConfiguration.DEFAULT_CONFIG_FILE_NAME)
        manager.safe_load_json_file(package_config)

    # If a user configuration file exists, load it
    if DefaultConfiguration.USER_CONFIGURATION_FILE:

        # Get full path
        if os.path.isabs(DefaultConfiguration.USER_CONFIGURATION_FILE):
            user_config_path = DefaultConfiguration.USER_CONFIGURATION_FILE
        elif "darwin" in sys.platform:
            user_config_path = os.path.join(os.getenv("HOME"), DefaultConfiguration.USER_CONFIGURATION_FILE)
        elif sys.platform == "win32":
            user_config_path = os.path.join(os.environ['USERPROFILE'], DefaultConfiguration.USER_CONFIGURATION_FILE)
        else:
            user_config_path = os.path.join("/etc", DefaultConfiguration.USER_CONFIGURATION_FILE)

        manager.safe_load_json_file(user_config_path)

    # If an app configuration file exists, load it
    if DefaultConfiguration.LOAD_FROM_APP_FOLDER:
        app_path = os.path.dirname(os.path.abspath(sys.argv[0]))
        if app_path == package_path:
            return

        if DefaultConfiguration.IGNORE_FILE and os.path.exists(os.path.join(app_path, DefaultConfiguration.IGNORE_FILE)):
            return

        app_config = os.path.join(app_path, DefaultConfiguration.DEFAULT_CONFIG_FILE_NAME)
        manager.safe_load_json_file(app_config)


_load_config_file()



"""This module is in charge of managing the rook's module state.

The external interface for the module is Rook located in rook.rook."""

import os
import struct

from rook.lib.logger import logger

from rook.interfaces.config import AgentAddress

from .augs.augs_manager import AugsManager
from .output import Output
from .com.agent_com import AgentCom
from .trigger_services import TriggerServices


class _Singleton(object):
    """This is singleton is the class managing the module.

    It should never be referred to directly, instead use obj in this module."""

    def __init__(self):
        """Initialize the object, sets member variables."""

        logger.info("Initializing Rook under process-%d", os.getpid())

        self._services_started = False

        self._trigger_services = TriggerServices()
        self._output = Output(self._trigger_services)
        self._aug_manager = None

        self._agent_com = None

    def _start_trigger_services(self):
        """Start trigger services.

        Calling this method multiple times will have no effect.
        """

        # Don't double init services
        if self._services_started:
            return

        self._trigger_services.start()
        self._aug_manager = AugsManager(self._trigger_services, self._output)
        self._services_started = True

    def _stop_trigger_services(self):
        if not self._services_started:
            return

        self._aug_manager = None
        self._trigger_services.close()

        self._services_started = False

    def set_rook_information(self, **kwargs):
        """Set the user-supplied version information for the application."""
        if self._agent_com:
            raise RookInterfaceException("Rook information must be set before connection!")

        self._output.set_user_info(kwargs)

    def collect_data(self, description, send_by_default, **kwargs):
        """Collect data based on a user triggered event."""
        raise NotImplementedError()

    def connect(self, host, port, token):
        """Connect to the Agent."""
        if self._agent_com:
            raise RookInterfaceException("Multiple connection attempts not supported!")

        if not host:
            try:
                host = os.environ['ROOKOUT_AGENT_HOST']
            except KeyError:
                host = AgentAddress.HOST

        if not port:
            try:
                port = os.environ['ROOKOUT_AGENT_PORT']
            except KeyError:
                port = AgentAddress.PORT

        if not token:
            token = os.environ.get('ROOKOUT_TOKEN')

        self._start_trigger_services()

        logger.debug("Initiating AgentCom-\t%s:%d", host, port)
        self._agent_com = AgentCom(self._aug_manager, self._trigger_services, self._output, host, port, token)
        self._output.set_agent_com(self._agent_com)

        self._agent_com.connect_to_agent()
        logger.info("Successful connection to agent")

singleton_obj = _Singleton()



import logging
import six
import inspect
import os
import threading

from rook.interfaces.config import LoggingServiceConfig

from rook.lib.logger import logger

from rook.lib.processor.namespaces.container_namespace import ContainerNamespace
from rook.lib.processor.namespaces.frame_namespace import FrameNamespace
from rook.augs.processor_extensions.namespaces.log_record_namespace import LogRecordNamespace


class LoggingLocationService(object):

    NAME = "logging"

    class Handler(logging.Handler):

        def __init__(self, name):
            super(LoggingLocationService.Handler, self).__init__()

            self._name = name

            self._logger = logging.getLogger(name)
            self._old_remove_handler = self._logger.removeHandler

            # Do not allow us to be removed
            def removeHandler(handler):
                if handler != self:
                    self._old_remove_handler(handler)
            self._logger.removeHandler = removeHandler

            self._logger.addHandler(self)

            self._augs = {}
            self._lock = threading.RLock()

        def close(self):
            self._logger.removeHandler = self._old_remove_handler
            self._logger.removeHandler(self)

        def emit(self, record):
            try:
                frame = FrameNamespace(self._get_frame(record))
                extracted = ContainerNamespace({'log_record': LogRecordNamespace(record)})

                with self._lock:
                    for aug in six.itervalues(self._augs):
                        aug.execute(frame, extracted)
            except:
                logger.exception("Error while processing log record")

        def add_aug(self, aug):
            with self._lock:
                self._augs[aug.aug_id] = aug

                aug.set_active()

        def remove_aug(self, aug_id):
            with self._lock:
                try:
                    aug = self._augs[aug_id]
                except KeyError:
                    return

                del self._augs[aug_id]
                aug.set_removed()

        def clear_augs(self):
            with self._lock:
                aug_ids = list(self._augs.keys())

                for aug_id in aug_ids:
                    self.remove_aug(aug_id)

        def empty(self):
            return len(self._augs) == 0

        def _get_frame(self, record):
            # Skip the top two frames (ours)
            frame = inspect.currentframe().f_back.f_back

            while hasattr(frame, "f_code"):
                filename = os.path.normcase(frame.f_code.co_filename)
                if filename == logging._srcfile:
                    frame = frame.f_back
                else:
                    return frame

            return None

    def __init__(self):
        if LoggingServiceConfig.BASIC_CONFIG_ROOT and len(logging.root.handlers) == 0:
            logging.basicConfig()

        self._handlers = dict()
        self._lock = threading.RLock()

    def add_logging_aug(self, logger, aug):
        with self._lock:
            try:
                handler = self._handlers[logger]
            except KeyError:
                handler = self.Handler(logger)
                self._handlers[logger] = handler

        handler.add_aug(aug)

    def remove_aug(self, aug_id):
        with self._lock:
            for handler in six.itervalues(self._handlers):
                handler.remove_aug(aug_id)

        self.prune_handlers()

    def clear_augs(self):
        with self._lock:
            for handler in six.itervalues(self._handlers):
                handler.clear_augs()

        self.prune_handlers()

    def prune_handlers(self):
        with self._lock:
            loggers_to_prune = [logger for logger, handler in six.iteritems(self._handlers) if handler.empty()]

            for logger in loggers_to_prune:
                self._handlers[logger].close()
                del self._handlers[logger]

    def close(self):
        self.clear_augs()





class ModuleProxy(object):

    def __init__(self, name):
        self.name = name
        self.module = None

    def _load_module(self):
        if not self.module:
            import importlib
            self.module = importlib.import_module(self.name, __package__)

    def __getattr__(self, item):
        self._load_module()
        return getattr(self.module, item)

rook_pb2 = ModuleProxy('.python.rook_pb2')
rook_pb2_grpc = ModuleProxy('.python.rook_pb2_grpc')



import logging

from .logging_types import LogType


class KWLogger(logging.LoggerAdapter):

    def __init__(self, name, extra_tags=None):
        self.extra_tags = extra_tags or []
        self.warn = self.warning
        return super(KWLogger, self).__init__(name, {})

    def __getattr__(self, item):
        # in case we don't have a function, delagate it to the logger
        return getattr(self.logger, item)

    def process(self, msg, kwargs):
        return msg, self.__kwargs_to_rookout_extra(msg, kwargs)

    def __kwargs_to_rookout_extra(self, msg, kwargs):
        rookout_extra = {"extraTags": self.extra_tags} if self.extra_tags else {}
        for k, v in kwargs.items():
            if k == "extra":
                rookout_extra.update(v)
            else:
                field_type = LogType.get_type(k)
                if field_type:
                    try:
                        rookout_extra[k] = field_type(v)
                    except Exception as e:
                        rookout_extra["kw_errors"] = rookout_extra.get("kw_errors", []) + [
                            "Error convert field {}".format(v)]
                else:
                    rookout_extra["kw_errors"] = rookout_extra.get("kw_errors", []) + [
                        "Missing Kw field {}".format(v)]

        if not rookout_extra:
            return {}
        rookout_extra["message"] = msg
        return {"extra": {"rookout_extra": rookout_extra}}



from rook.interface import Rook

def wrapper(f):
    def handler(event, context):
        ret = f(event, context)
        Rook().flush()
        return ret

    from rook import auto_start
    return handler


import six
import traceback


def rook_test():
    six.print_("\nTesting connection to agent\n")

    from rook.interfaces.config import LoggingConfiguration
    LoggingConfiguration.LOG_LEVEL = "DEBUG"
    LoggingConfiguration.LOG_TO_STDERR = False

    try:
        from rook import Rook
        connector = Rook()
        connector.start()
        success = True
    except:
        traceback.print_exc()
        success = False

    if success:
        six.print_("\n\n\nTest Finished Successfully!\n\n\n")
        exit(0)
    else:
        six.print_("\n\n\nTest Failed!\n\n\n")
        exit(1)

if '__main__' == __name__:
    rook_test()



from concurrent.futures.thread import ThreadPoolExecutor

from lib.logger import logger


class LimitedThreadPoolExecutor(ThreadPoolExecutor):
    def __init__(self, max_workers, max_items):
        super(LimitedThreadPoolExecutor, self).__init__(max_workers)
        self._max_items = max_items

    def submit(self, fn, *args, **kwargs):
        if self._work_queue.qsize() >= self._max_items:
            logger.warning("Discarding processing item")
            return

        return super(LimitedThreadPoolExecutor, self).submit(fn, *args, **kwargs)



if "rook" in __name__:
    # This is caused by not changing sys.path in Rook combined with the way we package the code
    from rook.interfaces.protobuf import rook_pb2
    from rook.interfaces.config import *
    from rook.interfaces.exceptions import *
else:
    from interfaces.protobuf import rook_pb2
    from interfaces.config import *
    from interfaces.exceptions import *
